package POWER::Event::IO::DNS;

use version; $VERSION = qv('0.2.0');    # update POD & Changes & README

use warnings;
use strict;
use Carp;

use Scalar::Util qw( dualvar weaken );

use POWER::Event::IO::share;
use POWER::Event::IO::UDP_CONNECTED;
use POWER::Utils;


our $TO_DNS
        = int 30 / $TICK_RATE + 0.5;# default timeout for receiving dns reply
our $TO_DNS_RESEND
        = int  2 / $TICK_RATE + 0.5;# default timeout for resending dns requests
my (@Timer_dns, @Timer_dns_resend);


#   ...
my $RE_IP   = qr{\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}}xms;

# WARNING    ...      
#             $DNS.     
#           DNS       ID...   @DNS.
my @DNS;                            # map DNS query ID => DNS query object

#   ,   inet_aton().
# WARNING      DNS- ( 
#           /etc/resolve.conf,    ).
my ($ns)    = `cat /etc/resolv.conf` =~ /^\s*nameserver\s+($RE_IP)\s*$/xms;
$DNS        = bless {
    ID          => 0,               # id for next DNS query
    }, __PACKAGE__;

sub _init {
    $DNS->{_io} = POWER::Event::IO::UDP_CONNECTED->new({
        host        => $ns || '127.0.0.1',
        port        => 53,
        wait_for    => $IN,
        cb          => $DNS,
    });
}


push @Tick_handlers, \&_tick;

sub _tick {
    #   .       
    # ӣ ݣ   $RESOLVING,   .
    for my $query (@{ shift @Timer_dns || [] }) {
        next unless $query;
        if (defined $query->{cb}) {
            run_callback(@{ $query }{'cb','method','host'}, undef,
                dualvar(-1, 'resolving timed out')
            );
        }
        #    ģ    $DNS->{ID}
        $DNS[ $query->{_id} ]  = undef;
    }
    #    DNS-. (  
    # ,  -   -   
    #   ...)      
    #    $RESOLVING,    
    #   .
    for my $query (@{ shift @Timer_dns_resend || [] }) {
        next unless $query;
        #    ,      - 
        # ݣ   $TO_DNS_RESEND... , ,  udp
        #       if    
        $DNS->{_io}->write($query->{_query})
            if !defined $DNS->{_io}->{out_buf};
        # -1  ..    tick_tack()  
        # (  $Timer_dns_resend[0]) ģ  
        #  
        unshift @{ $Timer_dns_resend[$TO_DNS_RESEND-1] }, $query;
        weaken     $Timer_dns_resend[$TO_DNS_RESEND-1][0];
    }
}

sub _resolve {
    my ($self, $host, $cb, $method) = @_;
    my $query = {
        host        => $host,
        cb          => $cb,
        method      => defined $method ? $method : 'RESOLVED',
        _id     => undef,               # $DNS->{ID} for this object
        _host   => undef,               # dns-packed {host}
        _query  => q{},                 # DNS query packet
    };
    # OPTIMIZATION    host     IP-, 
    #                . :)
    if ($query->{host} =~ /\A$RE_IP\z/xms) {
        run_callback(@{ $query }{'cb','method','host','host'});
        return;
    }
    if (length($query->{host}) >= 254) {
        run_callback(@{ $query }{'cb','method','host'}, undef,
            dualvar(-1, q{too long host name})
        );
        return;
    }
    weaken $query->{cb} if ref $query->{cb} && ref $query->{cb} ne 'CODE';
    #    DNS-:
    ($DNS->{ID} += 1) %= 0xFFFF+1;
    # WARNING           $query.
    #                cb,  
    #            ip   30 .  .
    $DNS[ $DNS->{ID} ]  = $query;
    $query->{_id} = $DNS->{ID};
    $query->{_host}
        = join(q{}, map {chr(length).$_} split /\./, lc($query->{host}))."\0";
    $query->{_query}
        = pack('nnnnnn a* nn',$DNS->{ID},0x0100,1,0,0,0,$query->{_host},1,1);
    $DNS->{_io}->write($query->{_query})
        if !defined $DNS->{_io}->{out_buf};
    unshift @{ $Timer_dns[$TO_DNS] }, $query;
    weaken     $Timer_dns[$TO_DNS][0];
    unshift @{ $Timer_dns_resend[$TO_DNS_RESEND] }, $query;
    weaken     $Timer_dns_resend[$TO_DNS_RESEND][0];
    return;
}

#    DNS-:
{ my $d = qr/[^\x00\xc0-\xff]*(?:\x00|[\xc0-\xff].)/xms;
sub IO {
    my ($self, $io) = @_;
    die "Oops. DNS resolver just died: $io->{err}" if $io->{events} & $ERR;
    my ($id,$flags,undef,$an) = unpack 'nnnn', my $pkt = delete $io->{in_buf};
    #   ""  -   ID (,
    #          
    #   $TO_DNS_RESEND)    ID   host
    # (  DNS-,  ID-   "").
    return if !$DNS[$id];
    return if $pkt !~ s/^.....\x01......\Q$DNS[$id]{_host}\E.\x01..//xms;
    my $query = $DNS[$id];
    #  - ݣ  ... :)
    if (defined $query->{cb}) {
        #   .
        if ($flags &= 0x0f) {
            run_callback(@{ $query }{'cb','method','host'}, undef,
                dualvar(-1, 'dns error: '.($flags==3 ? 'NXDOMAIN' : 'SRVERROR'))
            );
        }
        elsif (!(   ($pkt =~ s/^[^\x00\xC0-\xFF]*(?:\x00|[\xC0-\xFF].).\x01........//s) ||
                    (($an > ($pkt =~ s/\G$d.(?:[\x02\x05\x0C].{8}$d|\x06.{8}$d$d.{20}|\x0F.{8}..$d)//sog)) &&
                    ($pkt =~ s/^$d.\x01.{8}//so))  )) {
            run_callback(@{ $query }{'cb','method','host'}, undef,
                dualvar(-1, 'dns error: NODATA')
            );
        }
        # IP  ...
        else {
            run_callback(@{ $query }{'cb','method','host'},
                sprintf('%d.%d.%d.%d', unpack 'CCCC', $pkt)
            );
        }
    }
    $DNS[$id] = undef;
    return;
}}


1; # Magic true value required at end of module
__END__

=head1 DESCRIPTION

See L<POWER::Event::IO>.
