package POWER::Event::IO::SSL;

use version; $VERSION = qv('0.3.2');    # update POD & Changes & README

use warnings;
use strict;
use Carp;

use Scalar::Util qw( dualvar weaken );
use Socket;
use Crypt::MatrixSSL 1.83;

use base qw( POWER::Event::IO::plugin );
use POWER::Event::IO::share;


matrixSslOpen()
    == 0 or die 'matrixSslOpen';
END { matrixSslClose() }


our $TO_SSL_HANDSHAKE
        = int 30 / $TICK_RATE + 0.5;# default timeout for ssl handshake
my @Timer_ssl_handshake;

push @Tick_handlers, \&_tick;

# WARNING      _err!    - , 
#              -  : connection reset by peer
#            ssl handshake timeout (). ,   ,  
#                .
sub _tick {
    for my $io (@{ shift @Timer_ssl_handshake || [] }) {
        next unless $io;
        next if $io->{_ssl_handshaked};
        $io->{_events}         |= $ERR;
        $io->{_err}             = dualvar(-1, q{ssl handshake timed out});
        $Ready{ $io->{fd} }     = $io;
    }
}

sub DESTROY {
    my ($self) = @_;
    #    MatrixSSL.
    matrixSslDeleteSession($self->{_ssl});
    matrixSslFreeKeys($self->{_ssl_keys});
}

sub ATTACH {
    my ($self, $io) = @_;
    #    handshake.
    unshift @{ $Timer_ssl_handshake[$TO_SSL_HANDSHAKE] }, $io;
    weaken     $Timer_ssl_handshake[$TO_SSL_HANDSHAKE][0];
}

sub READ {
    my ($self, $io) = @_;
    $self->_real_read($io);
    # Return on error to avoid overwriting {_err} by SSL errors.
    return 0 if $io->{_events} & $ERR;
    # Try SSL decoding only if something was read.
    return 0 if !($io->{_events} & $IN);
    my $n = 0;
    while (length $self->{in_buf} && !($io->{_events} & $ERR)) {
        my ($error, $alertLevel, $alertDescription);
        my $rc = matrixSslDecode($self->{_ssl}, $self->{in_buf},
            my $buf=q{}, $error, $alertLevel, $alertDescription);
        if ($rc == $SSL_PROCESS_DATA) {
            $n    += length $buf;
            $_[2] .= $buf;
        }
        else {
            $self->{out_buf} .= $buf;
            # $ERR here can be set by previous _real_write() call.
            $self->_real_write($io) if !($io->{_events} & $ERR);
            if ($rc == $SSL_SUCCESS || $rc == $SSL_SEND_RESPONSE) {
                if (!$self->{_ssl_handshaked}) {
                    if (matrixSslHandshakeIsComplete($self->{_ssl})) {
                        $self->{_ssl_handshaked} = 1;
                        unshift @WRITE, $io; # push {out_buf}, if any
                        weaken $WRITE[0];
                    }
                }
            }
            # WARNING    $SSL_ERROR  $SSL_ALERT  in_buf
            #     ,   
            #    a)    
            #  $SSL_ERROR  $SSL_ALERT; b)   
            #  EPOLLIN   .
            #    ..  ERROR  ALERT 
            #  NO_CERTIFICATE.  NO_CERTIFICATE  
            #      
            #    MatrixSSL.
            elsif ($rc == $SSL_ERROR) {
                $io->{_err}     = dualvar(-1,
                    "SSL_ERROR: $SSL_alertDescription{$error}");
                $io->{_events} |= $ERR;
                last;
            }
            elsif ($rc == $SSL_ALERT) {
                if ($alertLevel == $SSL_ALERT_LEVEL_WARNING
                        && $alertDescription == $SSL_ALERT_CLOSE_NOTIFY) {
                    # Workaround MatrixSSL bug: ALERT packet doesn't removed
                    # from {in_buf}, and next matrixSslDecode() on this {in_buf}
                    # return SSL_ERROR while CLOSE_NOTIFY alert shouldn't be
                    # error at all. :(
                    $self->{in_buf} = q{};
                }
                else {
                    $io->{_err}     = dualvar(-1,
                        "SSL_ALERT: $SSL_alertLevel{$alertLevel}: ".
                        $SSL_alertDescription{$alertDescription}
    .(length($self->{in_buf}) ? ' [DEBUG: DATA IN in_buf]' : '')
                        );
                    $io->{_events} |= $ERR;
                }
                last;
            }
            elsif ($rc == $SSL_PARTIAL) {
                last;
            }
            else {
                die "matrixSslDecode: unexpected return code ($rc)\n";
            }
        }
    }
    return $n;
}

sub WRITE {
    my ($self, $io, undef, $out_pos) = @_;
    my $n = 0;
    if ($self->{_ssl_handshaked}) {
        my $s = substr($_[2], $out_pos);
        $n = length $s;
        while (length $s) {
            my $s2 = substr($s, 0, $SSL_MAX_PLAINTEXT_LEN, q{});
            matrixSslEncode($self->{_ssl}, $s2, $self->{out_buf})
                >= 0 or die 'matrixSslEncode';
        }
    }
    $self->_real_write($io);
    return $n;
}


1; # Magic true value required at end of module
__END__

=head1 DESCRIPTION

See L<POWER::Event::IO>.
