#######################
###   ###
#######################
#  :
# * STREAM
# * SSL::CLIENT
# * SSL::SERVER
# * PROXY::HTTPS
#   :
# * PROXY::HTTP
# * PROXY::SOCKS4
# * PROXY::SOCKS5
# * ޣ  (  )
# * /  I/O / 
# *  TCP  HTTP
# *  HTTP/1.1 Keep-Alive
#
###########
### API ###
###########
# $io->_read();
# $io->_write();
# $plugin->_real_read($io);
# $plugin->_real_write($io);
# $n = $plugin->{__next}->READ($io, $plugin->{in_buf});
# $n = $plugin->{__next}->WRITE($io, $plugin->{out_buf}, $plugin->{out_pos});
# unshift @READ, $io;  weaken $READ[0];
# unshift @WRITE, $io; weaken $WRITE[0];
# $DNS->_resolve($host, $plugin);
#
##############
### Events ###
##############
# $ERR   READ()/WRITE().
# _real_*()   $ERR  {err}=$EINBUFLIMIT    $ERR.
#     _real_*()  READ()/WRITE()  {err}==$EINBUFLIMIT, 
#   die() ..      READ()/WRITE().
# 
# $IN      $n ݣ  READ().
# 
# $EOF  is_eof   $Stream->READ().
#  $EOF   $IN.
# 
# $OUT      $n ݣ  WRITE().
# 
# $SENT       :
# {__wait_SENT} -    _real_write() 
#                   {out_buf}/{out_pos}   
#                   _real_write() $SENT
# {_has_out}    -      _real_write(), 
#                   _real_write()   
#                 WRITE()    {out_buf}/{out_pos}
# $SENT = {__wait_SENT} && !{_has_out};
#
################################
#    ###
################################
# 1)      $ERR    
#     .
# 2)    in_buf_limit      
#    _events  err.
# 3)   $n   >= 0.
# 4) WRITE()   _real_write(),    
# EPOLLOUT. READ()   _real_read(),    EPOLLIN.
#      ,      _real_* 
#    epoll_mod().
# 5)  READ()  WRITE()   __ _real_*   - 
#  $ERR.
# 6)     READ()       
#     READ(),   push(@READ, $io) 
#        (  EPOLLIN  ).
#     WRITE()        
# ,         (. 
#   handshake)    push(@WRITE, $io).
# 7)        READ()/WRITE()
# (, .  )      $io,    
#     ATTACH()    weaken!
#
###############################################
###   ͣ     ###
###############################################
# * /^[A-Z]/      - 
# * /^[a-z]/      -   
# * /^_[A-Za-z]/  -   
# * /^__/         -      -   :
#                   __wait_SENT       _real_write()
#                   __next            _real_write()  _real_read()
#                                    detach()  attach()


package POWER::Event::IO::plugin;

use version; $VERSION = qv('0.2.0');    # update POD & Changes & README

use warnings;
use strict;
use Carp;

use Scalar::Util qw( weaken dualvar );

use POWER::Event::IO::share;


sub new {
    my ($class, $opt_ref) = @_;
    my $self = bless {
        # must have:
        in_buf_limit=> undef,
        in_buf      => q{},
        in_bytes    => 0,
        out_pos     => undef,
        out_buf     => q{},
        out_bytes   => 0,
        }, $class;
    return $self;
}

sub ATTACH {
}

sub DETACH {
}

sub CONFIG {
    my ($self, $io) = @_;
    $self->_real_config($io);
    return;
}

sub READ {
    my ($self, $io) = @_;
    $self->{in_buf}     = q{};
    $self->{in_bytes}   = 0;
    $self->_real_read($io);
    $_[2] .= $self->{in_buf};
    return $self->{in_bytes};
}

sub WRITE {
    my ($self, $io, undef, $out_pos) = @_;
    $self->{out_buf}    = substr($_[2], $out_pos);
    $self->{out_bytes}  = 0;
    $self->_real_write($io);
    return $self->{out_bytes};
}

sub _real_config {
    my ($plugin, $io) = @_;
    $plugin->{__next}->CONFIG($io);
    return;
}

sub _real_read {
    my ($plugin, $io) = @_;
    #     in_buf_limit.  
    #        ޣ
    # in_buf_limit,     .   in_buf_limit -
    #   DoS ,     
    # in_buf,        .
    if ($plugin->{in_buf_limit}
            && length $plugin->{in_buf} > $plugin->{in_buf_limit}) {
        $io->{_err}           = $EINBUFLIMIT;
        $io->{_events}       |= $ERR;
        unshift @READ, $io;
        weaken $READ[0];
    }
    else {
        my $n = $plugin->{__next}->READ($io, $plugin->{in_buf});
        #   : $ERR, $EOF  $EAGAIN.
        if (!$n) {
            $io->{_events}      &= ~$IN;
        }
        #  .
        else {
            $io->{_events}      |= $IN;
            $plugin->{in_bytes} += $n;
        }
        # die 'Internal error' if $io->{_events} & $ERR && $io->{_err} == $EINBUFLIMIT;
    }
    return;
}

sub _real_write {
    my ($plugin, $io) = @_;
    if (defined $plugin->{out_pos}) {
        if ($plugin->{out_pos} < length $plugin->{out_buf}) {
            $plugin->{__wait_SENT} = 1;
        }
    }
    else {
        if ($plugin->{out_buf} ne q{}) {
            $plugin->{__wait_SENT} = 1;
        }
    }
    my $n = $plugin->{__next}->WRITE($io, $plugin->{out_buf},
        $plugin->{out_pos}||0);
    #   : $ERR, $EAGAIN    . :)
    if (!$n) {
        $io->{_events}         &= ~$OUT;
        if (defined $plugin->{out_pos}) {
            if ($plugin->{out_pos} < length $plugin->{out_buf}) {
                $io->{_events} &= ~$SENT;
                $io->{_has_out} = 1;
            }
        }
        else {
            if ($plugin->{out_buf} ne q{}) {
                $io->{_events} &= ~$SENT;
                $io->{_has_out} = 1;
            }
        }
    }
    #  .
    else {
        $io->{_events}         |= $OUT;
        $plugin->{out_bytes}   += $n;
        if (defined $plugin->{out_pos}) {
            $plugin->{out_pos} += $n;
            if ($plugin->{out_pos} < length $plugin->{out_buf}) {
                $io->{_has_out} = 1;
            }
            # die 'Internal error' if $plugin->{out_pos} > length($plugin->{out_buf});
        }
        else {
            substr($plugin->{out_buf}, 0, $n, q{});
            if ($plugin->{out_buf} ne q{}) {
                $io->{_has_out} = 1;
            }
        }
        if ($plugin->{__wait_SENT} && !$io->{_has_out}) {
            $plugin->{__wait_SENT} = 0;
            $io->{_events}     |= $SENT;
        }
        else {
            $io->{_events}     &= ~$SENT;
        }
    }
    return;
}


1; # Magic true value required at end of module
__END__

=head1 DESCRIPTION

See L<POWER::Event::IO>.

