use warnings;
use strict;
use Test::More;
use Test::Exception;
use Test::MockModule;

use File::Temp qw( tempfile );

use POSIX qw(locale_h);
setlocale(LC_ALL, 'C');

use POWER::Epoll;
use POWER::Event;
use POWER::Event::IO qw( :ALL );
use POWER::Utils::IO;

if ($INC{'Devel/Cover.pm'}) {
    plan tests => 54 + 2;
} else {
    plan tests => 54;
}

my ($io, $fh);

#      fd   $DNS
my ($dns) = `ls -al /proc/$$/fd/ | grep socket | head -n 1`=~/(\d+) -> socket/;

throws_ok { POWER::Event::IO->new()         } qr/Usage: POWER::Event::IO->new/;
throws_ok { POWER::Event::IO->new(undef)    } qr/Usage: POWER::Event::IO->new/;
throws_ok { POWER::Event::IO->new(1)        } qr/Usage: POWER::Event::IO->new/;
throws_ok { POWER::Event::IO->new({},undef) } qr/Usage: POWER::Event::IO->new/;


throws_ok { POWER::Event::IO->new({})               } qr/{cb} required/;
throws_ok { POWER::Event::IO->new({'cb'=>undef})    } qr/{cb} required/;


throws_ok { POWER::Event::IO->new({'cb'=>sub{}})
                                            } qr/{fh} or {host}\+{port}/;
throws_ok { POWER::Event::IO->new({'cb'=>sub{},fh=>undef})
                                            } qr/{fh} or {host}\+{port}/;
throws_ok { POWER::Event::IO->new({'cb'=>sub{},host=>undef})
                                            } qr/{fh} or {host}\+{port}/;
throws_ok { POWER::Event::IO->new({'cb'=>sub{},host=>undef,port=>undef})
                                            } qr/{fh} or {host}\+{port}/;
throws_ok { POWER::Event::IO->new({'cb'=>sub{},host=>1})
                                            } qr/{fh} or {host}\+{port}/;


throws_ok { POWER::Event::IO->new({'cb'=>sub{},fh=>1})
                                            } qr/{fh} is not open FILEHANDLE/;

open $fh, '| cat' or die "open: $!";
lives_ok  { $io=POWER::Event::IO->new({'cb'=>sub{},fh=>$fh}) };
throws_ok { POWER::Event::IO->new({'cb'=>sub{},fh=>$fh})
                                            } qr/epoll_add: File exists/;
throws_ok { POWER::Event::IO->new({'cb'=>sub{},fh=>$fh})
                                            } qr/epoll_add: File exists/;
close $fh;
throws_ok { POWER::Event::IO->new({'cb'=>sub{},fh=>$fh})
                                            } qr/{fh} is not open FILEHANDLE/;
open $fh, '| cat' or die "open: $!";
throws_ok { POWER::Event::IO->new({'cb'=>sub{},fh=>$fh})
                                            } qr/you must do \$io->close/;
throws_ok { POWER::Event::IO->new({'cb'=>sub{},fh=>$fh})
                                            } qr/epoll_add: File exists/;
$io->close();
throws_ok { POWER::Event::IO->new({'cb'=>sub{},fh=>$fh})
                                            } qr/{fh} is not open FILEHANDLE/;
open $fh, '| cat' or die "open: $!";
lives_ok  { POWER::Event::IO->new({'cb'=>sub{},fh=>$fh})->close() };
throws_ok { POWER::Event::IO->new({'cb'=>sub{},fh=>$fh})
                                            } qr/{fh} is not open FILEHANDLE/;


throws_ok { POWER::Event::IO::UDP->new({cb=>sub{},host=>'127.0.0.1',port=>1234})
                                            } qr/what to do.*IP for UDP/;
lives_ok  { $io=POWER::Event::IO::UDP->new({cb=>sub{},host=>'localhost',port=>1234}) };
throws_ok { loop()                          } qr/what to do.*IP for UDP/;
$io->close();


{
    my $module = Test::MockModule->new('POWER::Event::IO::UDP_CONNECTED', no_auto=>1);
    $module->mock('_real_read', sub {
        $module->original('_real_read')->(@_);
        $_[0]->{_events} |= $ERR;
        $_[0]->{_err}     = 'bad luck';
    });
    lives_ok { $io=POWER::Event::IO::UDP->new({cb=>sub{},host=>'localhost',port=>1234}) };
    throws_ok { loop() } qr/Oops. DNS resolver just died: bad luck/;
}


throws_ok { POWER::Event::IO::LISTEN->new({cb=>sub{},host=>'127.0.0.1',
    port=>1234,out_buf=>q{}})               } qr/Closed. Try again later/;
lives_ok { $io=POWER::Event::IO::LISTEN->new({cb=>sub{},host=>'127.0.0.1',
    port=>1234,out_buf=>undef})         };


throws_ok { $io->wait_for()             } qr/Usage: \$io->wait_for/;
throws_ok { $io->wait_for(1,2)          } qr/Usage: \$io->wait_for/;


throws_ok { $io->wait_for(undef)        } qr/wrong event mask/;
if ($INC{'Devel/Cover.pm'}) {
    #   warning
    throws_ok { $io->wait_for("ACCEPTED")   } qr/wrong event mask/;
}
throws_ok { $io->wait_for($CONNECTED)   } qr/wrong event mask/;
throws_ok { $io->wait_for($EOF)         } qr/wrong event mask/;
throws_ok { $io->wait_for($OUT)         } qr/wrong event mask/;
throws_ok { $io->wait_for($OUT|$ACCEPTED)}qr/wrong event mask/;
lives_ok  { $io->wait_for($ACCEPTED)    };
$io->close();


lives_ok  { $io=POWER::Event::IO->new({cb=>sub{},host=>'127.0.0.1',port=>1234}) };
throws_ok { $io->wait_for($EOF)         } qr/you must set {in_buf_limit}/;
lives_ok  { $io->wait_for($IN|$EOF)     };
throws_ok { $io->wait_for($EOF)         } qr/you must set {in_buf_limit}/;
$io->{in_buf_limit} = 0;
lives_ok  { $io->wait_for($EOF)         };


throws_ok { $io->write(1,2)             } qr/Usage: \$io->write/;
lives_ok  { $io->write(1)               };
lives_ok  { $io->write()                };
$io->close();


lives_ok  { $io=POWER::Event::IO::UDP_CONNECTED->new({cb=>sub{},
    host=>'127.0.0.1',port=>1234}) };
lives_ok  { $io->write(1)               };
lives_ok  { $io->write(2)               };
$io->{out_buf} = 3;
throws_ok { $io->write(4)               } qr/previous packet isn't sent yet/;


throws_ok { io_wait()                   } qr/usage: io_wait TIMEOUT/;
throws_ok { io_wait(1,2)                } qr/usage: io_wait TIMEOUT/;
if ($INC{'Devel/Cover.pm'}) {
    #  warning...
    throws_ok { io_wait("TIME OUT")         } qr/wrong TIMEOUT/;
}
throws_ok { io_wait(-1)                 } qr/wrong TIMEOUT/;
lives_ok  { io_wait(0)                  };


{
    my $module = Test::MockModule->new('POWER::Event::IO');
    $module->mock('epoll_wait', sub { return(1, [
        [ $dns, EPOLLERR ],
    ]) });
    lives_ok  { io_wait(0)              };
    $module->mock('epoll_wait', sub { return(1, [
        [ 1000, EPOLLERR ],
    ]) });
    throws_ok { io_wait(0)              } qr/no I\/O object.*fd=1000/;
}

