use warnings;
use strict;
use Test::More tests => 6;
use Test::Exception;
use Test::MockModule;

use POWER::Utils::Resource;
use POWER::Event;
use POWER::Event::IO;


my $max_mem_diff = $INC{'Devel/Cover.pm'} ? 10000 : 100;

leaktest('SSL_CLIENT', test=>1000, max_mem_diff=>2*$max_mem_diff); # 1.83 & 1.86 leaking here!?

leaktest('LISTEN', test=>10000, max_mem_diff=>$max_mem_diff);

my $io = POWER::Event::IO::UDP_CONNECTED->new({
        host        => '127.0.0.1',
        port        => 1234,
        cb          => sub {},
});
close($io->{fh});
leaktest('UDP_CONNECTED', test=>5000, max_mem_diff=>$max_mem_diff);


sub SSL_CLIENT {
    POWER::Event::IO->new({
        host        => '127.0.0.1',
        port        => 443,
        cb          => \&cb,
        plugin      => [
            ssl         => POWER::Event::IO::SSL::CLIENT->new({
#                trusted_CA  => '/etc/ssl/certs/ca-certificates.crt',
                cb_validate => \&cb,
            }),
        ],
    })->close();
#    @POWER::Event::IO::WRITE=();
#    @POWER::Event::IO::Timer_ssl_handshake=();
#    @POWER::Event::IO::Timer_connect=();
}

sub LISTEN {
    POWER::Event::IO::LISTEN->new({
        host        => '0.0.0.0',
        port        => 1234,
        cb          => \&cb,
        wait_for    => $ACCEPTED,
    })->close();
}

sub UDP_CONNECTED {
    eval {
        POWER::Event::IO::UDP_CONNECTED->new({
            host        => '127.0.0.1',
            port        => 1234,
            cb          => sub {},
        });
    };
    die "eval error: $@" if $@ !~ qr/you must do \$io->close/;
}

sub cb {}

sub leaktest {
    my $test = shift;
    my %arg  = (init=>10, test=>1000, max_mem_diff=>100, diag=>1, @_);
    my $code = do { no strict 'refs'; \&$test };
    $code->() for 1 .. $arg{init};
    my $mem = MEM_used();
    my $fd  = FD_used();
    $code->() for 1 .. $arg{test};
    diag sprintf "---- MEM\nWAS: %d\nNOW: %d\n", $mem, MEM_used() if $arg{diag};
    ok( abs(MEM_used() - $mem) < $arg{max_mem_diff},  "MEM: $test" );
    is( FD_used() - $fd, 0,                           " FD: $test" );
}

