use warnings;
use strict;
BEGIN { use lib '..'; require 't/.lib.t' }

my $banner  = "Server ready";
my $msg     = "Test message!\0\n";

@CheckPoint = (
    [ 'listener',   $ACCEPTED       ], 'accept incoming connection',
    [ 'ssl_client', $IN             ], 'client: got server banner',
    [ 'ssl_client', $banner         ], 'client: banner is correct',
    [ 'ssl_client', $SENT           ], 'client: request sent',
    [ 'ssl_server', $EOF            ], 'server: got eof',
    [ 'ssl_server', $msg            ], 'server: requst is correct',
    [ 'ssl_server', $SENT           ], 'server: reply sent',
    [ 'ssl_client', $EOF            ], 'client: got eof',
    [ 'ssl_client', "echo: $msg"    ], 'client: reply is correct',
);
plan tests => @CheckPoint/2;

POWER::Event::IO::LISTEN->new({
    host        => '127.0.0.1',
    port        => 4433,
    cb          => \&listener,
    wait_for    => $ACCEPTED,
});

POWER::Event::IO->new({
    host        => '127.0.0.1',
    port        => 4433,
    cb          => \&ssl_client,
    wait_for    => $IN,
    in_buf_limit=> 1024,
    plugin      => [
        ssl         => POWER::Event::IO::SSL::CLIENT->new({
            trusted_CA  => 't/cert/testca.crt',
        }),
    ],
});

eval { loop() };
die "out of loop: $@";

sub listener {
    my ($io) = @_;
    checkpoint($io->{events});
    POWER::Event::IO->new({
        fh          => delete $io->{in_buf},
        cb          => \&ssl_server,
        wait_for    => $EOF,
        in_buf_limit=> 1024,
        out_buf     => $banner,
        plugin      => [
            ssl         => POWER::Event::IO::SSL::SERVER->new({
                ssl_crt     => 't/cert/testsrv.crt',
                ssl_key     => 't/cert/testsrv.key',
            }),
        ],
    });
}

sub ssl_server {
    my ($io) = @_;
#  &diag_event;
    checkpoint($io->{events});
    if ($io->{events} & $ERR) {
        die $io->{err};
    }
    if ($io->{events} & $EOF) {
        checkpoint($io->{in_buf});
        $io->wait_for($EOF|$SENT);
        $io->write("echo: $io->{in_buf}");
    }
    if ($io->{events} & $SENT) {
        shutdown $io->{fh}, 1;
    }
}

sub ssl_client {
    my ($io) = @_;
#  &diag_event;
    checkpoint($io->{events});
    if ($io->{events} & $IN) {
        checkpoint($io->{in_buf});
        $io->{in_buf} = q{};
        $io->wait_for($SENT|$EOF);
        $io->write($msg);
    }
    if ($io->{events} & $SENT) {
        shutdown $io->{fh}, 1;
    }
    if ($io->{events} & $EOF) {
        checkpoint($io->{in_buf});
        exit;
    }
}

